#include "fgl.h"
/***
 Ferns' Graphics Library

 Routines om een GIF plaatje te laden.

 Behoorlijk simpele code die het Decode.C aanroept voor het eigenlijk
 decoden. Maakt gebruik van call-back funkties per scanline.

 (C)Copyright Ferns Paanakker, 1997.

 Inspired by code from Thomas G. Lane (code from IJG).
***/
#define	MAXCOLORMAPSIZE	256	/* max # of colors in a GIF colormap */
#define NUMCOLORS	3	/* # of colors */
#define CM_RED		0	/* color component numbers */
#define CM_GREEN	1
#define CM_BLUE		2

/* Macros for extracting header data --- note we assume chars may be signed */
#define LM_to_uint(a,b)		((((b)&0xFF) << 8) | ((a)&0xFF))
#define BitSet(byte, bit)	((byte) & (bit))
#define INTERLACE	0x40	/* mask for bit signifying interlaced image */
#define COLORMAPFLAG	0x80	/* mask for bit signifying colormap presence */

#define	ReadOK(file,buffer,len)	(_fsRead(file,buffer,len) == ((size_t) (len)))

static void BytesToNibbles(SHORT Width, BYTEP Source, BYTEP Dest);
static void BytesToBits(SHORT Width, BYTEP Source, BYTEP Dest);

static int is_interlaced;
static FHANDLE hFile;
static char buffer[4096];  // genoeg voor de meeste doeleinden...
static int data_precision;   /* bits of precision in image data */
static int iDataBytes=0;
static BYTE byData;
static PALETTE       MyPalette;   /* Belangrijk! */
static BMPINFOHEADER MyHeader;    /* Belangrijk! */

static int CurY=0;
int bad_code_count=0;
static SHORT rowbytes;
static ITEM  Data;

extern WORD decoder(WORD linewidth);

static SHORT Next_data()
{
 static BYTE fileBuf[4096];
 static SHORT index = 0;

 if (iDataBytes == 0)
 {
	if ((iDataBytes = _fsRead(hFile,&fileBuf[0],sizeof(fileBuf))) <= 0)
	 return FALSE;
	index = 0;
 }
 --iDataBytes;
 byData = *(fileBuf+(index++));

 return TRUE;
}

int get_byte()
{
  if (!Next_data()) bad_code_count++;
  return byData;
}


static int ReadByte ()
/* Read next byte from GIF file */
{
  BYTE c;

	if (_fsRead(hFile,&c,1) == -1)
		return 0;

  return (int)c;
}


static void ReadColorMap(int cmaplen)
/* Read a GIF colormap */
{
  register int i;

  for (i = 0; i < cmaplen; i++) {
    MyPalette.bmPalette[i].RGB.Red   = ReadByte();  // Red
    MyPalette.bmPalette[i].RGB.Green = ReadByte();  // Green
    MyPalette.bmPalette[i].RGB.Blue  = ReadByte();  // Blue
  }
}



/*
 * Read the file header; return image size and component count.
 *
 * Veel eruitgehaald. Leest geen Interlaced Images, het moet ook een
 * 256 kleuren plaatje zijn.
 *
 */

void start_input_gif ()
{
  char hdrbuf[10];		/* workspace for reading control blocks */
  unsigned int width, height;	/* image dimensions */
  int colormaplen, aspectRatio;
	int BitsPerPixel;
  int c;

	/* Kijk of de GIF header klopt */
  if (! ReadOK(hFile, hdrbuf, 6))
		return;
  if (hdrbuf[0] != 'G' || hdrbuf[1] != 'I' || hdrbuf[2] != 'F')
		return;
	/* Bekijk de versie nummers (maar genereer GEEN fatal error als de
	   check faalt). Geen consequentie -> dus disabled */
	/*
  if ((hdrbuf[3] != '8' || hdrbuf[4] != '7' || hdrbuf[5] != 'a') &&
      (hdrbuf[3] != '8' || hdrbuf[4] != '9' || hdrbuf[5] != 'a'))
	*/
	/* Lees Logical Screen Descriptor */
  if (! ReadOK(hFile, hdrbuf, 7))
		return;
  width = LM_to_uint(hdrbuf[0],hdrbuf[1]);
  height = LM_to_uint(hdrbuf[2],hdrbuf[3]);
	BitsPerPixel = hdrbuf[4] & 0x07;
	colormaplen = 2 << (BitsPerPixel);
  /* we ignore the color resolution, sort flag, and background color index */
  aspectRatio = hdrbuf[6] & 0xFF;
	// Nonfatal...
  //if (aspectRatio != 0 && aspectRatio != 49)

  /* Read global colormap if header indicates it is present */
  if (BitSet(hdrbuf[4], COLORMAPFLAG))
    ReadColorMap(colormaplen);

  /* Scan until we reach start of desired image.
   * We don't currently support skipping images, but could add it easily.
   */
  for (;;) {
    c = ReadByte();

    if (c == ';')		/* GIF terminator?? */
			return;

		//if (c == '!') {		/* Extension */
    //  DoExtension();
    //  continue;
    //}

    if (c != ',') {		/* Not an image separator? */
			//MYWARNMS1(JWRN_GIF_CHAR, c);
      continue;
    }

    /* Read and decipher Local Image Descriptor */
    if (! ReadOK(hFile, hdrbuf, 9))
			return;
    /* we ignore top/left position info, also sort flag */
    width = LM_to_uint(hdrbuf[4],hdrbuf[5]);
    height = LM_to_uint(hdrbuf[6],hdrbuf[7]);
    is_interlaced = BitSet(hdrbuf[8], INTERLACE);

    /* Read local colormap if header indicates it is present */
    /* Note: if we wanted to support skipping images, */
    /* we'd need to skip rather than read colormap for ignored images */
    if (BitSet(hdrbuf[8], COLORMAPFLAG)) {
      colormaplen = 2 << (hdrbuf[8] & 0x07);
      ReadColorMap(colormaplen);
    }
		/* // decoder.c doet dit al...
    input_code_size = ReadByte(); // get min-code-size byte
    if (input_code_size < 2 || input_code_size >= MAX_LZW_BITS)
      MYERREXIT1(JERR_GIF_CODESIZE, input_code_size);
		*/
    /* Reached desired image, so break out of loop */
    /* If we wanted to skip this image, */
    /* we'd call SkipDataBlocks and then continue the loop */
    break;
  }

  /* Return info about the image. */
	/* Vul ook de BMPINFOHEADER structuur */
  MyHeader.bmWidth    = width;
  MyHeader.bmHeight   = height;

  MyHeader.bmBitCount = 1;

	if (BitsPerPixel > 1)      /* Tussen 2 en 4 bits */
   MyHeader.bmBitCount = 4;
	if (BitsPerPixel > 4)
   MyHeader.bmBitCount = 8;  /* Tussen 5 en 8 bits */

  MyHeader.bmPlanes   = 1;                /* Kan kan niet werken met planes */
  MyHeader.bmXPelsPerMeter = 300;
  MyHeader.bmYPelsPerMeter = 300;

}

/*
 * Leest een GIF plaatje uit, voegt informatie toe aan het plaatje en
 * schrijft een nieuwe GIF weg.
 */
CLIPPER FGLLoadGif(void)
{
 BYTEP         bestand;
 ITEM          BitMap,ItemTemp;

 if (PCOUNT != 1)                // geef alleen naam door...
	{ _ret(); return; }

 iDataBytes=0;  // reset onze input-buffer...
 ItemTemp = _itemParam(1);            // bestand MOET zo geopend worden :-(
 if (_itemType(ItemTemp) != CHARACTER)
	{ _ret(); _itemRelease(ItemTemp); return; }
 bestand = _itemGetC(ItemTemp);
 _itemRelease(ItemTemp);
 hFile = _fsOpen(bestand,FO_READ);
 _itemFreeC(bestand);

 if (_fsError() )
	{ _ret(); return; }


 // lees header van de input_file en val de variabelen
 start_input_gif();

  BitMap = _itemArrayNew((WORD)7);  // maak array!!!!!!

  ItemTemp = _itemPutC(NULL,"DIB");
	_itemArrayPut(BitMap,1,ItemTemp);
	_itemRelease(ItemTemp);

  ItemTemp = _itemPutCL(NULL,(BYTEP)&MyHeader,sizeof(BMPINFOHEADER));
  _itemArrayPut(BitMap,2,ItemTemp);
	_itemRelease(ItemTemp);
  ItemTemp = _itemPutCL(NULL,(BYTEP)&MyPalette,sizeof(PALETTE));
  _itemArrayPut(BitMap,3,ItemTemp);
	_itemRelease(ItemTemp);

  Data = _itemArrayNew((WORD)MyHeader.bmHeight);  // maak array voor de Data.

  ItemTemp = _itemNew(NULL);
  _itemArrayPut(BitMap,5,ItemTemp);
	_itemRelease(ItemTemp);

  ItemTemp = _itemPutNL(NULL,(LONG)MyHeader.bmWidth);
  _itemArrayPut(BitMap,6,ItemTemp);
	_itemRelease(ItemTemp);

  ItemTemp = _itemPutNL(NULL,(LONG)MyHeader.bmHeight);
  _itemArrayPut(BitMap,7,ItemTemp);
	_itemRelease(ItemTemp);

 // Bereken nog even de rowbytes...
 rowbytes = (WORD)(((MyHeader.bmPlanes*MyHeader.bmBitCount*MyHeader.bmWidth)+31)/32)*4;

 // Okidoki. Ga het bestand decoden..
 CurY = 0;
 decoder((WORD)MyHeader.bmWidth);

 // voeg data toe aan de DIB
 BitMap = _itemArrayPut(BitMap,(WORD)4,Data);
 _itemReturn(BitMap);  // geef de bitmap terug in een array...
 _itemRelease(BitMap);
 _fsClose( hFile );
}


/* Voeg hier de logica toe om een regel te bewaren in DIB formaat */
int out_line(unsigned char *pixels,int linelen)
{
 BYTE buffie[2048]; /* Zeker voldoende */
 ITEM ItemTemp;

 switch(MyHeader.bmBitCount)
 {
	case 4    :
						 BytesToNibbles(linelen,&pixels[0],&buffie[0]);
						 break;
	case 1    :
						 BytesToBits(linelen,&pixels[0],&buffie[0]);
						 break;
	default   :
             memcpy(&buffie[0],&pixels[0],rowbytes);
 }

 ItemTemp = _itemPutCL(NULL,buffie,rowbytes);
 Data = _itemArrayPut(Data,(WORD)CurY+1,ItemTemp);
 _itemRelease(ItemTemp);
 CurY++;

 return(0);
}

static void BytesToBits(SHORT Width, BYTEP Source, BYTEP Dest)
{
 SHORT i;
 BYTE  mask;
 SHORT plek;


 mask = 0x80;
 plek = 0x00;
 Dest[plek] = 0;
 for (i=0;i<Width;i++)
 {
  if (Source[i] & 0x01)
   Dest[plek] |= mask;
  mask >>= 1;
  if (mask == 0x00)
  {
   mask = 0x80;
   plek++;
   Dest[plek] = 0x00;
  }
 }
}

static void BytesToNibbles(SHORT Width, BYTEP Source, BYTEP Dest)
{
 SHORT i;
 SHORT plek = 0;

 for (i=0;i<Width;i++)
 {
  if (i & 0x01)  /* Oneven (dus rechterkant vullen) */
  {
   Dest[plek] |= (Source[i] & 0x0F);
   plek++;
  }
  else
   Dest[plek] = (Source[i] << 4);
 }
}
